"""mouse_events module for EWOK.

Copyright (C) 2022 Theodore B. Ruegsegger
See LICENSE file for terms of use.
"""

# from global_constants import *
# from design import *
from animation import *

def animate_click(design, canvas, event):
    ##################################################################
    # Why?
    design.redraw_last_figure(canvas)
    ##################################################################
    mouse_xy = mouse_xy_uc(event, design.pixels_per_unit,
                           design.origin_cc)
    design.current_figure = design.closest(mouse_xy)
    if design.current_figure is None:
        return
    figure_list = design.current_figure.figure_list
    linkage_type = design_type(design, figure_list)
    if (figure_list is not None
        and not (linkage_type == "incomplete"
                 or linkage_type == "over-specified")):
        fig_index = design.current_figure.fig_index
        if figure_list == design.positions:
            pass # for now. Might eventually just grab a coupler
        elif figure_list == design.sliderpoints:
            design.animation = LinkageAnimation(design, fig_index,
                                                linkage_type)
            angle = design.sliderpoints[fig_index].angle
            design.animation.redraw(design, canvas, angle, mouse_xy,
                                    design.pixels_per_unit,
                                    design.origin_cc)
            design.highlight_current_figure(canvas)
            design.mode = "Animate"
        elif figure_list == design.turning_blocks:
            pass # until I figure out how to animate turning blocks
        elif figure_list == design.circlepoints :
            design.animation = LinkageAnimation(design, fig_index,
                                                linkage_type)
            angle = direction(
                mouse_xy - design.centerpoints[fig_index].center_uc)
            if design.rotate_by_steps:
                angle = nearest_5_degrees(angle)
            design.animation.redraw(design, canvas, angle, mouse_xy,
                                    design.pixels_per_unit,
                                    design.origin_cc)
            design.highlight_current_figure(canvas)
            design.mode = "Animate"
    
def animate_move(design, canvas, event):
    if design.mode == "Animate":
        mouse_xy = mouse_xy_uc(event, design.pixels_per_unit,
                               design.origin_cc)
        fig_index = design.current_figure.fig_index
        if design.animation.linkage_type == "slider-crank" \
           or design.animation.linkage_type == "double-slider":
            angle = design.sliderpoints[fig_index].angle
        else:
            angle = direction(
                mouse_xy - design.centerpoints[fig_index].center_uc)
            if design.rotate_by_steps:
                angle = nearest_5_degrees(angle)
        design.animation.redraw(design, canvas, angle, mouse_xy,
                                design.pixels_per_unit,
                                design.origin_cc)

def animate_release(design, canvas, event):
    if design.mode == "Animate":
        design.animation.erase(canvas)
        design.update_all(canvas)
    design.mode = "Select"
    
def design_type(design, figure_list):
    k_count = len(design.circlepoints)
    m_count = len(design.centerpoints)
    s_count =  len(design.sliderpoints)
    if k_count == 2 and m_count == 2 and s_count == 0:
        return "double-crank"
    elif k_count == 1 and m_count == 1 and s_count == 1:
        if figure_list == design.circlepoints:
            return "crank-slider"
        else:
            return "slider-crank"
    elif k_count == 0 and m_count == 0 and s_count == 2:
        return "double-slider"
    elif k_count == m_count and k_count + s_count > 2:
        return "over-specified"
    else:
        return "incomplete"

def leave_canvas(design, canvas, task, event):
    if task == "Design" or task == "Animate":
        design.redraw_last_figure(canvas)

def left_click(design, canvas, event):
    mouse_xy = mouse_xy_uc(event, design.pixels_per_unit,
                           design.origin_cc)
    if design.mode == "AddPosn":
        fig_index = len(design.positions)
        tag = "Axis"+str(fig_index)
        if fig_index == 0:
            color = DESIGN_POSITION_COLOR
            what = "Design position"
        else:
            color = ALT_POSITION_COLOR
            what = "Alternate position"
        design.positions.append(Axis(mouse_xy, 0, tag, color))
        design.current_figure = SelectedItem(
            design.positions, fig_index, design.positions[fig_index])
        design.highlight_current_figure(canvas)
        report_figure_status(canvas, what, mouse_xy, 0)
        design.mode = "MovePosn"
    elif design.mode == "AddCirPt":
        fig_index = len(design.circlepoints)
        tag = "K"+str(fig_index)
        design.circlepoints.append(
            CirclePoint(design.coupler_coordinates(mouse_xy, 0), tag,
                        DESIGN_POSITION_COLOR))
        design.move_circlepoint(canvas, mouse_xy, fig_index)
        what = "Circlepoint"
        report_figure_status(
            canvas, what, design.circlepoints[fig_index].center_uc, 0)
        design.mode = "MoveCirPt"
    elif design.mode == "AddSlider":
        fig_index = design.create_sliderpoint(canvas, mouse_xy)
        what = "Sliderpoint"
        report_figure_status(
            canvas, what, design.sliderpoints[fig_index].center_uc,
            design.current_figure.the_figure.angle)
    elif design.mode == "AddCtrPt":
        fig_index = len(design.centerpoints)
        tag = "M"+str(fig_index)
        design.centerpoints.append(
            CenterPoint(mouse_xy, tag, FIXED_PLANE_COLOR))
        design.move_centerpoint(canvas, mouse_xy, fig_index)
        what = "Centerpoint"
        report_figure_status(
            canvas, what, design.centerpoints[fig_index].center_uc, 0)
        design.mode = "MoveCtrPt"
    elif design.mode == "AddTurnBlk":
        fig_index = len(design.turning_blocks)
        tag = "C"+str(fig_index)
        design.turning_blocks.append(
            TurningBlock( mouse_xy, 0, tag, FIXED_PLANE_COLOR))
        design.move_turning_block(canvas, mouse_xy, fig_index)
        what = "Turning block"
        report_figure_status(canvas, what, mouse_xy, 0)
        design.mode = "MoveTurnBlk"
    elif design.mode == "Select":
        nearest = design.closest(mouse_xy)
        if design.current_figure != nearest:
            design.redraw_last_figure(canvas)
            design.current_figure = nearest
        if nearest is None:
            pass
        elif nearest.figure_list == design.centerpoints:
            design.move_centerpoint(canvas, mouse_xy,
                                    nearest.fig_index)
            what = "Centerpoint"
            report_figure_status(canvas, what, mouse_xy, 0)
            design.mode = "MoveCtrPt"
        elif nearest.figure_list == design.circlepoints:
            design.move_circlepoint(canvas, mouse_xy,
                                    nearest.fig_index)
            what = "Circlepoint"
            report_figure_status(
                canvas, what,
                design.circlepoints[nearest.fig_index].center_uc, 0)
            design.mode = "MoveCirPt"
        elif nearest.figure_list == design.sliderpoints:
            design.move_slider(canvas, mouse_xy, nearest.fig_index)
            what = "Sliderpoint"
            report_figure_status(
                canvas, what,
                design.sliderpoints[nearest.fig_index].center_uc, 0)
            design.mode = "MoveSlider"
        elif nearest.figure_list == design.turning_blocks:
            design.move_turning_block(canvas, mouse_xy,
                                      nearest.fig_index)
            what = "Turning block"
            report_figure_status(canvas, what, mouse_xy, 0)
            design.mode = "MoveTurnBlk"
        elif nearest.figure_list == design.positions:
#            if (len(design.centerpoints)
#                + len(design.circlepoints)) > 0:
#                ## Handle position changes wrt Ms and Ks
#                print "Can't handle late spec changes yet"
            design.current_figure.the_figure.center_uc = mouse_xy
            design.highlight_current_figure(canvas)
            if design.current_figure.fig_index == 0:
                what = "Design position"
            else:
                what = "Alternate position"
            report_figure_status(
                canvas, what, mouse_xy,
                design.current_figure.the_figure.angle)
            design.mode = "MovePosn"
    elif design.mode == "DelFig":
        design.redraw_last_figure(canvas)
        nearest = design.closest(mouse_xy)
        # If it's a pivot pair, delete crank assembly
        if nearest.figure_list == design.centerpoints \
           and len(design.circlepoints) > nearest.fig_index \
           or nearest.figure_list == design.circlepoints \
           and len(design.centerpoints) > nearest.fig_index:
                design.delete_crank_assembly(canvas,
                                             nearest.fig_index)
        # If it's a position, check which one
        elif nearest.figure_list == design.positions:
            # Erase the position's image...
            nearest.the_figure.erase(canvas)
            # ...and delete the position from its list
            del nearest.figure_list[nearest.fig_index]
            # If any but the highest-numbered position
            if nearest.fig_index < len(design.positions):
                # Reorder positions: erase, retag, recolor, redraw
                for i in range(len(design.positions)):
                    design.positions[i].erase(canvas)
                    design.positions[i].tag = "Axis"+str(i)
                    if i == 0:
                        design.positions[i].color \
                            = DESIGN_POSITION_COLOR
                    else:
                        design.positions[i].color \
                            = ALT_POSITION_COLOR
                    design.positions[i].render(
                        canvas, design.positions[i].center_uc,
                        design.positions[i].angle,
                        design.pixels_per_unit, design.origin_cc,
                        design.positions[i].color)
            # If no positions left, delete all circlepoints,
            # sliderpoints, and links
            if len(design.positions) == 0:
                for i in range(len(design.circlepoints)):
                    design.circlepoints[i].erase(canvas)
                    canvas.delete("Crank"+str(i))
                design.circlepoints=[]
                for i in range(len(design.sliderpoints)):
                    design.sliderpoints[i].erase(canvas)
                design.sliderpoints=[]
            ## Redraw any unpaired circlepoints
            ## Redraw all cranks
        else:
            # It's an unpaired centerpoint or circlepoint,
            # or a sliderpoint or a concurrency point
            # Erase the object's image(s), ...
            nearest.the_figure.erase(canvas)
            # ...and delete the object from its list
            del nearest.figure_list[nearest.fig_index]
        canvas.configure(cursor=DEFAULT_CURSOR)
        if (len(design.positions) + len(design.circlepoints)
            + len(design.centerpoints)) > 0:
            design.mode = "Select"
            design.current_figure = design.closest(mouse_xy)
            design.highlight_current_figure(canvas)
        else:
            design.mode = "Init"

def left_move(design, canvas, event):
    mouse_xy = mouse_xy_uc(event, design.pixels_per_unit,
                           design.origin_cc)
    if design.mode == "MovePosn":
        design.current_figure.the_figure.center_uc = mouse_xy
        design.highlight_current_figure(canvas)
        if design.current_figure.fig_index == 0:
            what = "Design position"
        else:
            what = "Alternate position"
        report_figure_status(canvas, what, mouse_xy,
                             design.current_figure.the_figure.angle)
    elif design.mode == "MoveCirPt":
        design.move_circlepoint(canvas, mouse_xy,
                                design.current_figure.fig_index)
        what = "Circlepoint"
        report_figure_status(
            canvas, what,
            design.current_figure.the_figure.center_uc, 0)
    elif design.mode == "MoveSlider":
        design.move_slider(canvas, mouse_xy,
                           design.current_figure.fig_index)
        what = "Sliderpoint"
        report_figure_status(
            canvas, what, design.current_figure.the_figure.center_uc,
            design.current_figure.the_figure.angle)
    elif design.mode == "MoveCtrPt":
        design.move_centerpoint(canvas, mouse_xy,
                                design.current_figure.fig_index)
        what = "Centerpoint"
        report_figure_status(
            canvas, what,
            design.current_figure.the_figure.center_uc, 0)
    elif design.mode == "MoveTurnBlk":
        design.move_turning_block(canvas, mouse_xy,
                                  design.current_figure.fig_index)
        what = "Turning block"
        report_figure_status(canvas, what, mouse_xy,
                             design.current_figure.the_figure.angle)
    # Redraw coupler frame, if enabled
    if design.show_frame:
        design.create_coupler_frame()
        design.redraw_coupler_images(canvas, design.coupler_border,
                                     DESIGN_POSITION_COLOR)
    # Redraw coupler sketches, if any
    for sketch in design.coupler_sketches:
        design.redraw_coupler_images(canvas, sketch,
                                     DESIGN_POSITION_COLOR)
    # Redraw fixed-frame sketches, if any
    for sketch in design.fixed_sketches:
        sketch.redraw(canvas, 0+0j, 0, design.pixels_per_unit,
                      design.origin_cc, FIXED_PLANE_COLOR)

def left_release(design, canvas, event):
    left_move(design, canvas, event)
    # Erase any displayed loci or poles
    canvas.delete("locus")
    for i in range(len(design.poles)):
        canvas.delete("Pole"+str(i))
    canvas.delete("PoleCircle")
    if design.mode != "Init":
        design.mode = "Select"

# Can't assume there's a middle button on non-X platforms
def middle_click(design, canvas, event):
    pass

def middle_move(design, canvas, event):
    pass

def middle_release(design, canvas, event):
    pass

def mouse_motion(design, canvas, event):
    mouse_xy = mouse_xy_uc(event, design.pixels_per_unit,
                           design.origin_cc)
    if design.mode == "Select" or design.mode == "DelFig":
        nearest = design.closest(mouse_xy)
        if design.current_figure != nearest:
            design.redraw_last_figure(canvas)
            design.current_figure = nearest
            design.highlight_current_figure(canvas)

def mouse_xy_uc(event, pixels_per_unit, origin_cc):
    # return user_coordinates(complex(event.x, event.y))
    return user_coordinates(complex(event.x, event.y),
                            pixels_per_unit, origin_cc)

def nearest_5_degrees(angle):
    return round(angle/RADIANS_5_DEGREES, 0) * RADIANS_5_DEGREES

def report_figure_status(canvas, what, center_uc, angle=0):
    def display(coordinate):
        return str(round(coordinate, 3))
    
    where = "(" + display(center_uc.real) + "," \
        + display(center_uc.imag) +")  "
    tilt = str(round(angle*DEGREES_PER_RADIAN, 2)) + "°"
    canvas.statusbar.update(text=what+": "+where+tilt)

def right_click(design, canvas, event):
    if design.mode == "Select":
        design.redraw_last_figure(canvas)
        mouse_xy = mouse_xy_uc(event, design.pixels_per_unit,
                               design.origin_cc)
        design.current_figure = design.closest(mouse_xy)
        if design.current_figure.figure_list is None:
            pass
        elif design.current_figure.figure_list == design.positions:
            angle = direction(
                mouse_xy - design.current_figure.the_figure.center_uc)
            if design.rotate_by_steps:
                angle = nearest_5_degrees(angle)
            design.current_figure.the_figure.angle = angle
            design.highlight_current_figure(canvas)
            if design.current_figure.fig_index == 0:
                what = "Design position"
            else:
                what = "Alternate position"
            report_figure_status(
                canvas, what,
                design.current_figure.the_figure.center_uc,
                design.current_figure.the_figure.angle)
            # Redraw coupler frame, if enabled
            if design.show_frame:
                design.create_coupler_frame()
                design.redraw_coupler_images(
                    canvas, design.coupler_border,
                    DESIGN_POSITION_COLOR)
            # Redraw coupler sketches, if any
            for sketch in design.coupler_sketches:
                design.redraw_coupler_images(canvas, sketch,
                                             DESIGN_POSITION_COLOR)
            # Redraw fixed-frame sketches, if any
            for sketch in design.fixed_sketches:
                sketch.redraw(canvas, 0+0j, 0, design.pixels_per_unit,
                              design.origin_cc, FIXED_PLANE_COLOR)
            design.mode = "TurnFig"
        elif design.current_figure.figure_list == design.sliderpoints:
            # Can't rotate the sliderpoint directly because it will
            # jump to the K-locus (with 2 or 3 coupler positions), so
            # create a dial (rotating arrow) to set the sliding
            # direction
            abs_center_uc = design.absolute_coordinates(
                design.current_figure.the_figure.center_uc, 0)
            angle = direction(mouse_xy - abs_center_uc)
            if design.rotate_by_steps:
                angle = nearest_5_degrees(angle)
            design.dial = Dial(abs_center_uc, angle, "Dial",
                               DESIGN_POSITION_COLOR)
            design.dial.redraw(canvas, abs_center_uc, angle,
                               design.pixels_per_unit,
                               design.origin_cc,
                               DESIGN_POSITION_COLOR)
            design.mode = "TurnSlider"
            design.current_figure.the_figure.angle = angle
            design.highlight_current_figure(canvas)
            what = "Sliderpoint"
            report_figure_status(
                canvas, what,
                design.current_figure.the_figure.center_uc,
                design.current_figure.the_figure.angle)
        elif design.current_figure.figure_list \
             == design.turning_blocks:
            ## Very dirty trick to rotate turning blocks for
            ## screenshot (angle is normally determined by geometry)
            angle = direction(
                mouse_xy - design.current_figure.the_figure.center_uc)
            if design.rotate_by_steps:
                angle = nearest_5_degrees(angle)
            design.current_figure.the_figure.angle = angle
            design.highlight_current_figure(canvas)
            what = "Turning block"
            report_figure_status(
                canvas, what,
                design.current_figure.the_figure.center_uc,
                design.current_figure.the_figure.angle)
            design.mode = "TurnFig"
        ## End of very dirty trick

def right_move(design, canvas, event):
    if design.mode == "TurnFig":
        mouse_xy = mouse_xy_uc(event, design.pixels_per_unit,
                               design.origin_cc)
        angle = direction(
            mouse_xy - design.current_figure.the_figure.center_uc)
        if design.rotate_by_steps:
            angle = nearest_5_degrees(angle)
        design.current_figure.the_figure.angle = angle
        figure_list = design.current_figure.figure_list
        if figure_list == design.positions:
            if design.current_figure.fig_index == 0:
                what = "Design position"
            else:
                what = "Alternate position"
        elif figure_list == design.circlepoints:
            what = "Circlepoint"
        elif figure_list == design.centerpoints:
            what = "Centerpoint"
        elif figure_list == design.sliderpoints:
            what = "Sliderpoint"
        elif figure_list == design.turning_blocks:
            what = "Turning block"
        report_figure_status(
            canvas, what, design.current_figure.the_figure.center_uc,
            design.current_figure.the_figure.angle)
        design.highlight_current_figure(canvas)
        # Redraw coupler frame, if enabled
        if design.show_frame:
            design.create_coupler_frame()
            design.redraw_coupler_images(canvas,
                                         design.coupler_border,
                                         DESIGN_POSITION_COLOR)
        # Redraw coupler sketches, if any
        for sketch in design.coupler_sketches:
            design.redraw_coupler_images(canvas, sketch,
                                         DESIGN_POSITION_COLOR)
        # Redraw fixed-frame sketches, if any
        for sketch in design.fixed_sketches:
            sketch.redraw(canvas, 0+0j, 0, design.pixels_per_unit,
                          design.origin_cc, FIXED_PLANE_COLOR)
    elif design.mode == "TurnSlider":
        mouse_xy = mouse_xy_uc(event, design.pixels_per_unit,
                               design.origin_cc)
        angle = direction(mouse_xy - design.dial.center_uc)
        if design.rotate_by_steps:
            angle = nearest_5_degrees(angle)
        design.dial.angle = angle
        design.dial.redraw(canvas, design.dial.center_uc, angle,
                           design.pixels_per_unit, design.origin_cc,
                           DESIGN_POSITION_COLOR)
        design.current_figure.the_figure.angle = angle
        design.turn_slider(canvas, mouse_xy,
                           design.current_figure.fig_index)
        what = "Sliderpoint"
        report_figure_status(
            canvas, what,
            design.current_figure.the_figure.center_uc,
            design.current_figure.the_figure.angle)
        # Redraw coupler frame, if enabled
        if design.show_frame:
            design.create_coupler_frame()
            design.redraw_coupler_images(
                canvas, design.coupler_border, DESIGN_POSITION_COLOR)
        # Redraw coupler sketches, if any
        for sketch in design.coupler_sketches:
            design.redraw_coupler_images(canvas, sketch,
                                         DESIGN_POSITION_COLOR)
        # Redraw fixed-frame sketches, if any
        for sketch in design.fixed_sketches:
            sketch.redraw(canvas, 0+0j, 0, design.pixels_per_unit,
                          design.origin_cc, FIXED_PLANE_COLOR)

def right_release(design, canvas, event):
    if design.mode == "TurnFig":
        right_move(design, canvas, event)
    elif design.mode == "TurnSlider":
        right_move(design, canvas, event)
        design.dial.erase(canvas)
        # Erase any displayed loci or poles
        canvas.delete("locus")
        for i in range(len(design.poles)):
            canvas.delete("Pole"+str(i))
            canvas.delete("PoleCircle")
    design.mode = "Select"

def sketch_click(design, canvas, sketch_mode, task, event):
    mouse_xy = mouse_xy_uc(event, design.pixels_per_unit,
                           design.origin_cc)
    if sketch_mode is None:
        pass
    elif sketch_mode == "Add Sketch":
        # Either start a new polyline or continue the current one
        # (design.current_sketch)
        if design.current_sketch is None:
            # Starting a new one; create it and draw a point (circle)
            design.current_sketch = Polyline_uu([mouse_xy])
            sketch_redraw_current(design, canvas)
        else:
            # Continuing; add a point and draw the polyline
            design.current_sketch.points.append(mouse_xy)
            sketch_redraw_current(design, canvas)
    elif sketch_mode == "Delete Sketch":
        sketches = sketch_list(design, task) # by reference
        nearest = sketch_nearest(sketches, mouse_xy)
        if nearest is not None:
            canvas.delete(sketches[nearest].tag)
            del sketches[nearest]
        sketch_resume(design, canvas, task)
    report_figure_status(canvas, "Cursor", mouse_xy)

def sketch_click_double(design, canvas, sketch_mode, task, event):
    # Conclude sketching current polyline, add it to
    # design.coupler_sketches as a polygon, then set
    # design.current_sketch to None
    if sketch_mode is None:
        pass
    elif sketch_mode == "Add Sketch":
        # Add latest point and close the polygon
        mouse_xy = mouse_xy_uc(event, design.pixels_per_unit,
                               design.origin_cc)
        design.current_sketch.points.append(mouse_xy)
        sketch_redraw_current(design, canvas)
        sketches = sketch_list(design, task) # by reference
        sketches.append(Sketch(0+0j, design.current_sketch.points,
                               sketch_tag(task, len(sketches)),
                               sketch_color(task, True)))
        canvas.delete("CurrentSketch")
        design.current_sketch = None
        canvas.configure(cursor=DEFAULT_CURSOR)
        sketch_resume(design, canvas, task)
    elif sketch_mode == "Delete Sketch":
        pass

def sketch_click_right(design, canvas, sketch_mode, event):
    # Delete last point of current polyline
    if sketch_mode is None:
        pass
    elif sketch_mode == "Add Sketch":
        if design.current_sketch is None:
            pass
        elif len(design.current_sketch.points) > 0:
            design.current_sketch.points.pop()
            if len(design.current_sketch.points) == 0:
                design.current_sketch = None
            sketch_redraw_current(design, canvas)

def sketch_color(task, final=False):
    if task == "Sketch Coupler":
        if final:
            return DESIGN_POSITION_COLOR
        else:
            return SKETCH_COUPLER_TASK_COLOR
    elif task == "Sketch Fixed":
        if final:
            return FIXED_PLANE_COLOR
        else:
            return SKETCH_FIXED_TASK_COLOR

def sketch_list(design, task):
    # sketch_list points by reference to actual list
    if task == "Sketch Coupler":
        return design.coupler_sketches
    elif task == "Sketch Fixed":
        return design.fixed_sketches

def sketch_mouse_motion(design, canvas, sketch_mode, task, event):
    # Mouse moves with no buttons clicked.
    mouse_xy = mouse_xy_uc(event, design.pixels_per_unit,
                           design.origin_cc)
    if sketch_mode is None:
        pass
    elif sketch_mode == "Add Sketch":
        if design.current_sketch is not None:
            sketch_redraw_current(design, canvas)
    elif sketch_mode == "Delete Sketch":
        sketches = sketch_list(design, task) # by reference
        nearest = sketch_nearest(sketches, mouse_xy)
        for i in range(len(sketches)):
            if i == nearest:
                color = HIGHLIGHT_COLOR
            else:
                color = sketch_color(task)
            sketches[i].render(canvas, 0+0j, 0,
                               design.pixels_per_unit,
                               design.origin_cc, color)
    report_figure_status(canvas, "Cursor", mouse_xy)

def sketch_move(design, canvas, sketch_mode, event):
    # Mouse moves with left button pressed. Adjust position of last
    # point of current polygon
    mouse_xy = mouse_xy_uc(event, design.pixels_per_unit,
                           design.origin_cc)
    if sketch_mode is None:
        pass
    elif sketch_mode == "Add Sketch":
        design.current_sketch.points[-1] = mouse_xy
        sketch_redraw_current(design, canvas)
    elif sketch_mode == "Delete Sketch":
        pass
    report_figure_status(canvas, "Cursor", mouse_xy)

def sketch_move_right(design, canvas, sketch_mode, event):
    # Mouse moves with right button pressed. Adjust position of last
    # point of current polygon (after removing the one after it by
    # clicking)
    mouse_xy = mouse_xy_uc(event, design.pixels_per_unit,
                           design.origin_cc)
    if sketch_mode is None:
        pass
    elif sketch_mode == "Add Sketch":
        if design.current_sketch is None:
            pass
        else:
            design.current_sketch.points[-1] = mouse_xy
            sketch_redraw_current(design, canvas)
    
def sketch_nearest(sketches, mouse_xy):
    min_distance = float("INFINITY")
    nearest = None
    for i in range(len(sketches)):
        for point in sketches[i].geometry[0].points:
            distance = abs(point - mouse_xy)
            if distance <= min_distance:
                min_distance = distance
                index_nearest = i
    return index_nearest

def sketch_redraw_current(design, canvas):
    canvas.delete("CurrentSketch")
    if design.current_sketch is None:
        pass
    elif len(design.current_sketch.points) > 0:
        # Draw up to the last point
        first = canvas_coordinates(design.current_sketch.points[0],
                                   design.pixels_per_unit,
                                   design.origin_cc)
        if len(design.current_sketch.points) == 1:
            # Only one point so far; draw a point (circle)
            canvas.create_oval(first.real-2, first.imag-2,
                               first.real+2, first.imag+2,
                               tags="CurrentSketch",
                               outline=HIGHLIGHT_COLOR)
        else:
            design.current_sketch.draw(canvas, 0+0j, 0,
                                       design.pixels_per_unit,
                                       design.origin_cc,
                                       "CurrentSketch",
                                       HIGHLIGHT_COLOR)
        # Connect last point to first with dashed line
        last = canvas_coordinates(design.current_sketch.points[-1],
                                  design.pixels_per_unit,
                                  design.origin_cc)
        canvas.create_line(last.real, last.imag, first.real,
                           first.imag, tags="CurrentSketch",
                           fill=HIGHLIGHT_COLOR, dash=(4))
    
def sketch_resume(design, canvas, task):
    canvas.delete("CurrentSketch")
    if task == "Sketch Coupler":
        sketch_resume_coupler(design, canvas)
    elif task == "Sketch Fixed":
        sketch_resume_fixed(design, canvas)
    if design.current_sketch is not None:
        canvas.delete("CurrentSketch")
        design.current_sketch.draw(canvas, 0+0j, 0,
                                   design.pixels_per_unit,
                                   design.origin_cc, "CurrentSketch",
                                   HIGHLIGHT_COLOR)

def sketch_resume_coupler(design, canvas):
    for sketch in design.coupler_sketches:
        sketch.render(canvas, 0+0j, 0, design.pixels_per_unit,
                      design.origin_cc, SKETCH_COUPLER_TASK_COLOR)
    for circlepoint in design.circlepoints:
        circlepoint.erase(canvas)
        circlepoint.render(canvas, circlepoint.center_uc, 0,
                           design.pixels_per_unit, design.origin_cc,
                           DESIGN_POSITION_COLOR)
    for sliderpoint in design.sliderpoints:
        sliderpoint.erase(canvas)
        sliderpoint.render(canvas, sliderpoint.center_uc,
                           sliderpoint.angle, design.pixels_per_unit,
                           design.origin_cc, DESIGN_POSITION_COLOR)
    for centerpoint in design.centerpoints:
        centerpoint.erase(canvas)
        center = design.coupler_coordinates(centerpoint.center_uc, 0)
        centerpoint.render(canvas, center, 0, design.pixels_per_unit,
                           design.origin_cc, FIXED_PLANE_COLOR)
    # Draw crank(s) if any
        m_count = len(design.centerpoints)
        k_count = len(design.circlepoints)
        for i in range(min(m_count, k_count)):
            m_ctr = design.coupler_coordinates(
                design.centerpoints[i].center_uc, 0)
            k_ctr = design.circlepoints[i].center_uc
            link = Link(m_ctr, k_ctr, design.pixels_per_unit,
                            "Link"+str(i), DESIGN_POSITION_COLOR)
            link.erase(canvas)
            link.render(canvas, link.center_uc,
                              link.angle, design.pixels_per_unit,
                              design.origin_cc, DESIGN_POSITION_COLOR)

def sketch_resume_fixed(design, canvas):
    design.update_all(canvas)
    for sketch in design.fixed_sketches:
        sketch.render(canvas, 0+0j, 0, design.pixels_per_unit,
                      design.origin_cc, SKETCH_FIXED_TASK_COLOR)

def sketch_tag(task, index):
    if task == "Sketch Coupler":
        suffix = "C"
    else:
        suffix = "F"
    return "Sketch_" + suffix + str(index)

